@@ -1,6 +1,6 @@
 package Lemonldap::NG::Handler::Main::Reload;
 
-our $VERSION = '2.0.6';
+our $VERSION = '2.0.9';
 
 package Lemonldap::NG::Handler::Main;
 
@@ -13,6 +13,7 @@ use Scalar::Util qw(weaken);
 
 use constant UNPROTECT => 1;
 use constant SKIP      => 2;
+use constant MAYSKIP   => 3;
 
 our @_onReload;
 
@@ -179,11 +180,13 @@ sub jailInit {
     my ( $class, $conf ) = @_;
 
     $class->tsv->{jail} = Lemonldap::NG::Handler::Main::Jail->new( {
-            useSafeJail     => $conf->{useSafeJail},
-            customFunctions => $conf->{customFunctions},
+            useSafeJail          => $conf->{useSafeJail},
+            customFunctions      => $conf->{customFunctions},
+            multiValuesSeparator => $conf->{multiValuesSeparator},
         }
     );
-    $class->tsv->{jail}->build_jail( $class, $conf->{require} );
+    $class->tsv->{jail}
+      ->build_jail( $class, $conf->{require}, $conf->{requireDontDie} );
 }
 
 ## @imethod protected void defaultValuesInit(hashRef args)
@@ -212,7 +215,7 @@ sub defaultValuesInit {
         # Override with vhost options
         if ( $conf->{vhostOptions} ) {
             my $name = 'vhost' . ucfirst($opt);
-            foreach my $vhost ( keys %{ $conf->{vhostOptions} } ) {
+            foreach my $vhost ( sort keys %{ $conf->{vhostOptions} } ) {
                 $conf->{vhostOptions}->{$vhost} ||= {};
                 my $val = $conf->{vhostOptions}->{$vhost}->{$name};
 
@@ -226,11 +229,13 @@ sub defaultValuesInit {
         }
     }
     if ( $conf->{vhostOptions} ) {
-        foreach my $vhost ( keys %{ $conf->{vhostOptions} } ) {
+        foreach my $vhost ( sort keys %{ $conf->{vhostOptions} } ) {
             $class->tsv->{type}->{$vhost} =
               $conf->{vhostOptions}->{$vhost}->{vhostType};
             $class->tsv->{authnLevel}->{$vhost} =
               $conf->{vhostOptions}->{$vhost}->{vhostAuthnLevel};
+            $class->tsv->{serviceTokenTTL}->{$vhost} =
+              $conf->{vhostOptions}->{$vhost}->{vhostServiceTokenTTL};
         }
     }
     return 1;
@@ -271,10 +276,30 @@ sub locationRulesInit {
     my ( $class, $conf, $orules ) = @_;
 
     $orules ||= $conf->{locationRules};
+    $class->tsv->{vhostReg} = [];
+    my @lastReg;
 
     foreach my $vhost ( keys %$orules ) {
         my $rules = $orules->{$vhost};
-        $class->tsv->{locationCount}->{$vhost} = 0;
+        if ( $vhost =~ /[\%\*]/ ) {
+            my $expr = join '[^\.]*', map {
+                my $elt = $_;
+                join '.*', map { quotemeta $_ } split /\*/, $elt;
+            } split /\%/, $vhost;
+            if ($expr) {
+                push @{ $class->tsv->{vhostReg} }, [ qr/^$expr$/, $vhost ];
+            }
+            else {
+                push @lastReg, [ qr/.+/, $vhost ];
+            }
+        }
+        $class->tsv->{locationCount}->{$vhost}         = 0;
+        $class->tsv->{locationCondition}->{$vhost}     = [];
+        $class->tsv->{locationProtection}->{$vhost}    = [];
+        $class->tsv->{locationRegexp}->{$vhost}        = [];
+        $class->tsv->{locationConditionText}->{$vhost} = [];
+        $class->tsv->{locationAuthnLevel}->{$vhost}    = [];
+
         foreach my $url ( sort keys %{$rules} ) {
             my ( $cond, $prot ) = $class->conditionSub( $rules->{$url} );
             unless ($cond) {
@@ -293,10 +318,14 @@ sub locationRulesInit {
                 push @{ $class->tsv->{locationCondition}->{$vhost} },  $cond;
                 push @{ $class->tsv->{locationProtection}->{$vhost} }, $prot;
                 push @{ $class->tsv->{locationRegexp}->{$vhost} },     qr/$url/;
+                push @{ $class->tsv->{locationAuthnLevel}->{$vhost} },
+                  $url =~ /\(\?#AuthnLevel=(-?\d+)\)/
+                  ? $1
+                  : undef;
                 push @{ $class->tsv->{locationConditionText}->{$vhost} },
-                    $cond =~ /^\(\?#(.*?)\)/ ? $1
-                  : $cond =~ /^(.*?)##(.+)$/ ? $2
-                  :                            $url;
+                    $url =~ /^\(\?#(.*?)\)/ ? $1
+                  : $url =~ /^(.*?)##(.+)$/ ? $2
+                  :                           $url;
                 $class->tsv->{locationCount}->{$vhost}++;
             }
         }
@@ -307,12 +336,22 @@ sub locationRulesInit {
             $class->tsv->{defaultProtection}->{$vhost} = 0;
         }
     }
+    @{ $class->tsv->{vhostReg} } = sort {
+        my $av = $a->[1];
+        my $bv = $b->[1];
+        return 1  if $av =~ /^\*/ and $bv !~ /^\*/;
+        return -1 if $bv =~ /^\*/ and $av !~ /^\*/;
+        return 1  if $av =~ /^\%/ and $bv !~ /^\%/;
+        return -1 if $bv =~ /^\%/ and $av !~ /^\%/;
+        return length($bv) <=> length($av) || $av cmp $bv;
+    } @{ $class->tsv->{vhostReg} } if @{ $class->tsv->{vhostReg} };
+    push @{ $class->tsv->{vhostReg} }, @lastReg if @lastReg;
     return 1;
 }
 
 ## @imethod protected void sessionStorageInit(hashRef args)
 # Initialize the Apache::Session::* module choosed to share user's variables
-# and the Cache::Cache module choosed to cache sessions
+# and the Cache::Cache module chosen to cache sessions
 # @param $args reference to the configuration hash
 sub sessionStorageInit {
     my ( $class, $conf ) = @_;
@@ -349,18 +388,19 @@ sub sessionStorageInit {
         if ( $conf->{status} ) {
             my $params = "";
             if ( $class->tsv->{sessionCacheModule} ) {
-                $params = ' ' . join(
+                $params = $class->tsv->{sessionCacheModule} . ',{' . join(
                     ',',
-                    $class->tsv->{sessionCacheModule} . map {
-                        "$_ => "
-                          . $class->tsv->{sessionCacheOptions}->{$_}
-                    } keys %{ $class->tsv->{sessionCacheOptions} // {} }
-                );
+                    map {
+                        "$_ => '"
+                          . $class->tsv->{sessionCacheOptions}->{$_} . "'"
+                      }
+                      keys %{ $class->tsv->{sessionCacheOptions} // {} }
+                ) . '}';
             }
             $class->tsv->{statusPipe}->print("RELOADCACHE $params\n");
         }
     }
-return 1;
+    return 1;
 }
 
 ## @imethod void headersInit(hashRef args)
@@ -444,6 +484,7 @@ sub postUrlInit {
 # @return array (ref(sub), int)
 sub conditionSub {
     my ( $class, $cond ) = @_;
+    $cond =~ s/\(\?#(\d+)\)$//;
     my ( $OK, $NOK ) = ( sub { 1 }, sub { 0 } );
 
     # Simple cases : accept and deny
@@ -512,9 +553,8 @@ sub conditionSub {
         eval 'use Apache2::Filter' unless ( $INC{"Apache2/Filter.pm"} );
         return (
             sub {
-
                 my ($req) = @_;
-                $class->localUnlog;
+                $class->localUnlog( $req, @_ );
                 $req->{env}->{'psgi.r'}->add_output_filter(
                     sub {
                         my $r = $_[0]->r;
@@ -532,6 +572,9 @@ sub conditionSub {
         );
     }
 
+    my $mayskip = 0;
+    $mayskip = MAYSKIP if $cond =~ /\bskip\b/;
+
     # Replace some strings in condition
     $cond = $class->substitute($cond);
     my $sub;
@@ -541,7 +584,7 @@ sub conditionSub {
     }
 
     # Return sub and protected flag
-    return ( $sub, 0 );
+    return ( $sub, $mayskip );
 }
 
 ## @method arrayref aliasInit
@@ -575,6 +618,10 @@ sub substitute {
     $expr =~ s/\$(?!(?:ENV|env)\b)(_\w+|[a-zA-Z]\w*)/\$s->{$1}/sg;
     $expr =~ s/\$ENV\{/\$r->{env}->\{/g;
     $expr =~ s/\$env->\{/\$r->{env}->\{/g;
+    $expr =~ s/\bskip\b/q\{999_SKIP\}/g;
+
+    # handle inGroup
+    $expr =~ s/\binGroup\(([^)]*)\)/listMatch(\$s->{'hGroups'},$1,1),/g;
 
     return $expr;
 }
